/*

*/

package renderer.pipeline;
import  renderer.scene.*;
import  renderer.framebuffer.*;

import java.util.List;
import java.util.ArrayList;

/**
   This renderer takes as its input a {@link Scene} data structure
   and a {@link FrameBuffer.Viewport} within a {@link FrameBuffer}
   data structure. This renderer mutates the {@link FrameBuffer.Viewport}
   so that it is filled in with the rendered image of the geometric
   scene represented by the {@link Scene} object.
<p>
   This implements our third rendering pipeline. It is essentially the
   same as the previous renderer but it adds an antialiasing option to
   the rasterizing stage, {@link RasterizeAntialias}. So there are
   still just three pipeline stages.
*/
public final class Pipeline
{
   public static boolean debug = false;

   /**
      Mutate the {@link FrameBuffer}'s default {@link FrameBuffer.Viewport}
      so that it holds the rendered image of the {@link Scene} object.

      @param scene  {@link Scene} object to render
      @param fb     {@link FrameBuffer} to hold rendered image of the {@link Scene}
   */
   public static void render(final Scene scene, final FrameBuffer fb)
   {
      render(scene, fb.vp); // render into the default viewport
   }


   /**
      Mutate the {@link FrameBuffer}'s given {@link FrameBuffer.Viewport}
      so that it holds the rendered image of the {@link Scene} object.

      @param scene  {@link Scene} object to render
      @param vp     {@link FrameBuffer.Viewport} to hold rendered image of the {@link Scene}
   */
   public static void render(final Scene scene, final FrameBuffer.Viewport vp)
   {
      // Render every Model in the Scene.
      for (final Model model : scene.modelList)
      {
         if ( model.visible )
         {
            logMessage(model, "==== Render Model: " + model.name + " ====");

            check(model);

            logVertexList("0. Camera   ", model);

            // 1. Apply the projection transformation.
            final Model model2 = Projection.project(model, scene.camera);

            logVertexList("1. Projected", model2);

            // 2. Clip line segments to the camera's view rectangle.
            final Model model3 = Clip.clip(model2);

            logVertexList("2. Clipped  ", model3);
            logLineSegmentList("2. Clipped  ", model3);

            // 3. Rasterize each visible line segment into pixels.
            for (final LineSegment ls : model3.lineSegmentList)
            {
               logLineSegment("3. Rasterize", model3, ls);

               RasterizeAntialias.rasterize(model3, ls, vp);
            }

            logMessage(model, "==== End Model: " + model.name + " ====");
         }
         else
         {
            logMessage(model, "==== Hidden model: " + model.name + " ====");
         }
      }
   }


   /**
      Determine if there are any obvious problems with the {@link Model}
      being rendered. The purpose of these checks is to make the renderer
      a bit more user friendly. If a user makes a simple mistake and tries
      to render a {@link Model} that is missing vertices or line segments,
      then the user gets a helpful error message.

      @param model  the {@link Model} being rendered
   */
   public static void check(final Model model)
   {
      boolean error = false;
      if (model.vertexList.isEmpty())
      {
         System.err.println("***WARNING: This model does not have any vertices.");
         error = true;
      }
      if (model.lineSegmentList.isEmpty())
      {
         System.err.println("***WARNING: This model does not have any line segments.");
         error = true;
      }
      if (error)
      {
         System.err.println(model);
      }
   }


   /**
      Use the pipeline's and the model's debug variables to determine
      if the given message should be printeed to stderr.

      @param model    the {@link Model} being rendered
      @param message  {@link String} to output to stderr
   */
   public static void logMessage(final Model model, final String message)
   {
      if (Pipeline.debug || model.debug)
         System.err.println( message );
   }


   /**
      This method prints a {@link String} representation of the given
      {@link Model}'s {@link Vertex} list.

      @param stage  name for the pipeline stage
      @param model  the {@link Model} whose vertex list is to be printed
   */
   private static void logVertexList(final String stage, final Model model)
   {
      if (Pipeline.debug || model.debug)
      {
         int i = 0;
         for (final Vertex v : model.vertexList)
         {
            System.err.printf( stage + ": index = %3d, " + v.toString(), i );
            ++i;
         }
      }
   }


   /**
      This method prints a {@link String} representation of the given
      {@link Model}'s {@link LineSegment} list.

      @param stage  name for the pipeline stage
      @param model  the {@link Model} whose line segment list is to be printed
   */
   private static void logLineSegmentList(final String stage, final Model model)
   {
      if (Pipeline.debug || model.debug)
      {
         for (final LineSegment ls : model.lineSegmentList)
         {
            System.err.printf( stage + ": " + ls.toString() );
         }
      }
   }


   /**
      This method prints a {@link String} representation of the given
      {@link LineSegment}.

      @param stage  name for the pipeline stage
      @param model  {@link Model} that the {@link LineSegment} {@code ls} comes from
      @param ls     {@link LineSegment} whose string representation is to be printed
   */
   public static void logLineSegment(final String stage, final Model model, final LineSegment ls)
   {
      if (Pipeline.debug || model.debug)
      {
         System.err.print( stage + ": " + ls.toString() );
         final int index0 = ls.index[0];
         final int index1 = ls.index[1];
         final Vertex v0 = model.vertexList.get(index0);
         final Vertex v1 = model.vertexList.get(index1);
         System.err.printf("   index = %3d, " + v0.toString(), index0);
         System.err.printf("   index = %3d, " + v1.toString(), index1);
      }
   }
}
